@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013,2015,2016 Olly Betts
  * Copyright 2008 Lemur Consulting Ltd
  * Copyright 2010 Richard Boulton
  *
@@ -28,6 +28,7 @@
 
 #include <xapian/error.h>
 
+#include "errno_to_string.h"
 #include "safeerrno.h"
 #ifdef __WIN32__
 # include "msvc_posix_wrapper.h"
@@ -222,7 +223,7 @@ FlintTable::read_block(uint4 n, byte * p) const
 	    if (errno == EBADF && handle == -2)
 		FlintTable::throw_database_closed();
 	    string message = "Error reading block " + str(n) + ": ";
-	    message += strerror(errno);
+	    errno_to_string(errno, message);
 	    throw Xapian::DatabaseError(message);
 	} else if (bytes_read == 0) {
 	    string message = "Error reading block " + str(n) + ": got end of file";
@@ -241,7 +242,7 @@ FlintTable::read_block(uint4 n, byte * p) const
 	if (errno == EBADF && handle == -2)
 	    FlintTable::throw_database_closed();
 	string message = "Error seeking to block: ";
-	message += strerror(errno);
+	errno_to_string(errno, message);
 	throw Xapian::DatabaseError(message);
     }
 
@@ -300,7 +301,7 @@ FlintTable::write_block(uint4 n, const byte * p) const
 	} else if (bytes_written == -1) {
 	    if (errno == EINTR) continue;
 	    string message = "Error writing block: ";
-	    message += strerror(errno);
+	    errno_to_string(errno, message);
 	    throw Xapian::DatabaseError(message);
 	} else if (bytes_written == 0) {
 	    string message = "Error writing block: wrote no data";
@@ -317,7 +318,7 @@ FlintTable::write_block(uint4 n, const byte * p) const
 #else
     if (lseek(handle, (off_t)block_size * n, SEEK_SET) == -1) {
 	string message = "Error seeking to block: ";
-	message += strerror(errno);
+	errno_to_string(errno, message);
 	throw Xapian::DatabaseError(message);
     }
 
@@ -566,7 +567,7 @@ FlintTable::split_root(uint4 split_n)
     /* check level overflow - this isn't something that should ever happen
      * but deserves more than an Assert()... */
     if (level == BTREE_CURSOR_LEVELS) {
-	throw Xapian::DatabaseCorruptError("Btree has grown impossibly large ("STRINGIZE(BTREE_CURSOR_LEVELS)" levels)");
+	throw Xapian::DatabaseCorruptError("Btree has grown impossibly large (" STRINGIZE(BTREE_CURSOR_LEVELS) " levels)");
     }
 
     byte * q = zeroed_new(block_size);
@@ -676,9 +677,11 @@ FlintTable::mid_point(byte * p)
 	}
     }
 
-    /* falling out of mid_point */
+    /* This shouldn't happen, as the sum of the item sizes should be the same
+     * as the value calculated in size, so assert but return a sane value just
+     * in case. */
     Assert(false);
-    return 0; /* Stop compiler complaining about end of method. */
+    return dir_end;
 }
 
 /** add_item_to_block(p, kt_, c) adds item kt_ to the block at p.
@@ -1438,6 +1441,11 @@ FlintTable::basic_open(bool revision_supplied, flint_revision_number_t revision_
 
     base_letter = ch;
 
+    if (cursor_created_since_last_modification) {
+	cursor_created_since_last_modification = false;
+	++cursor_version;
+    }
+
     /* ready to open the main file */
 
     return true;
@@ -1507,7 +1515,7 @@ FlintTable::do_open_to_write(bool revision_supplied,
 	string message(create_db ? "Couldn't create " : "Couldn't open ");
 	message += name;
 	message += "DB read/write: ";
-	message += strerror(errno);
+	errno_to_string(errno, message);
 	throw Xapian::DatabaseOpeningError(message);
     }
 
@@ -1853,14 +1861,6 @@ FlintTable::commit(flint_revision_number_t revision, int changes_fd,
 	    C[i].rewrite = false;
 	}
 
-	// Do this as late as possible to allow maximum time for writes to be
-	// committed.
-	if (!io_sync(handle)) {
-	    (void)::close(handle);
-	    handle = -1;
-	    throw Xapian::DatabaseError("Can't commit new revision - failed to flush DB to disk");
-	}
-
 	// Save to "<table>.tmp" and then rename to "<table>.base<letter>" so
 	// that a reader can't try to read a partially written base file.
 	string tmp = name;
@@ -1869,25 +1869,21 @@ FlintTable::commit(flint_revision_number_t revision, int changes_fd,
 	basefile += "base";
 	basefile += char(base_letter);
 	base.write_to_file(tmp, base_letter, tablename, changes_fd, changes_tail);
-#if defined __WIN32__
-	if (msvc_posix_rename(tmp.c_str(), basefile.c_str()) < 0)
-#else
-	if (rename(tmp.c_str(), basefile.c_str()) < 0)
-#endif
-	{
-	    // With NFS, rename() failing may just mean that the server crashed
-	    // after successfully renaming, but before reporting this, and then
-	    // the retried operation fails.  So we need to check if the source
-	    // file still exists, which we do by calling unlink(), since we want
-	    // to remove the temporary file anyway.
-	    int saved_errno = errno;
-	    if (unlink(tmp) == 0 || errno != ENOENT) {
-		string msg("Couldn't update base file ");
-		msg += basefile;
-		msg += ": ";
-		msg += strerror(saved_errno);
-		throw Xapian::DatabaseError(msg);
-	    }
+
+	// Do this as late as possible to allow maximum time for writes to
+	// happen, and so the calls to io_sync() are adjacent which may be
+	// more efficient, at least with some Linux kernel versions.
+	if (changes_tail ? !io_full_sync(handle) : !io_sync(handle)) {
+	    (void)::close(handle);
+	    handle = -1;
+	    (void)unlink(tmp);
+	    throw Xapian::DatabaseError("Can't commit new revision - failed to flush DB to disk");
+	}
+
+	if (!io_tmp_rename(tmp, basefile)) {
+	    string msg("Couldn't update base file ");
+	    msg += basefile;
+	    throw Xapian::DatabaseError(msg, errno);
 	}
 	base.commit();
 
@@ -1986,6 +1982,11 @@ FlintTable::cancel()
     changed_n = 0;
     changed_c = DIR_START;
     seq_count = SEQ_START_POINT;
+
+    if (cursor_created_since_last_modification) {
+	cursor_created_since_last_modification = false;
+	++cursor_version;
+    }
 }
 
 /************ B-tree reading ************/
@@ -2006,7 +2007,7 @@ FlintTable::do_open_to_read(bool revision_supplied, flint_revision_number_t revi
 	string message("Couldn't open ");
 	message += name;
 	message += "DB to read: ";
-	message += strerror(errno);
+	errno_to_string(errno, message);
 	throw Xapian::DatabaseOpeningError(message);
     }
 
@@ -2188,13 +2189,14 @@ FlintTable::prev_default(Cursor_ * C_, int j) const
 {
     byte * p = C_[j].p;
     int c = C_[j].c;
-    Assert(c >= DIR_START);
-    Assert((unsigned)c < block_size);
-    Assert(c <= DIR_END(p));
+    AssertRel(DIR_START,<=,c);
+    AssertRel(c,<,DIR_END(p));
+    AssertRel((unsigned)DIR_END(p),<=,block_size);
     if (c == DIR_START) {
 	if (j == level) return false;
 	if (!prev_default(C_, j + 1)) return false;
 	c = DIR_END(p);
+	AssertRel(DIR_START,<,c);
     }
     c -= D2;
     C_[j].c = c;
@@ -2209,9 +2211,14 @@ FlintTable::next_default(Cursor_ * C_, int j) const
 {
     byte * p = C_[j].p;
     int c = C_[j].c;
-    Assert(c >= DIR_START);
+    AssertRel(c,<,DIR_END(p));
+    AssertRel((unsigned)DIR_END(p),<=,block_size);
     c += D2;
-    Assert((unsigned)c < block_size);
+    if (j > 0) {
+	AssertRel(DIR_START,<,c);
+    } else {
+	AssertRel(DIR_START,<=,c);
+    }
     // Sometimes c can be DIR_END(p) + 2 here it appears...
     if (c >= DIR_END(p)) {
 	if (j == level) return false;
